import hashlib
import logging
import os
import sqlite3
import urlparse

import psycopg2
import psycopg2.extras

from bdi.interface0 import Interface0
from bdi.utils.aminoacids import seq_md5



localhost_db_url = 'postgresql://localhost/uta'
public_db_url = 'postgresql://uta_public:uta_public@uta.invitae.com/uta'
default_db_url = os.environ.get('UTA_DB_URL',public_db_url)
logger = logging.getLogger(__name__)

def connect(db_url=default_db_url):
    """
    Connect to a UTA database instance and return a UTA0 interface instance.

    When called with an explicit db_url argument, that db_url is used for connecting.

    When called without an explicit argument, the function default is
    determined by the environment variable UTA_DB_URL if it exists, or
    bdi.sources.uta0.public_db_url.

    The format of the db_url is driver://user:pass@host/database (the same
    as that used by SQLAlchemy).  Examples:

    A remote public postgresql database:
        postgresql://uta_public:uta_public@uta.invitae.com/uta'

    A local postgresql database:
        postgresql://localhost/uta
    
    A local SQLite database:
      sqlite:////tmp/uta-0.0.5.db
    """

    url = urlparse.urlparse(db_url)
    
    if url.scheme == 'sqlite':
        return UTA0_sqlite(url.path)

    if url.scheme == 'postgresql':
        return UTA0_postgresql(
            host = url.hostname,
            port = url.port or 5432,
            database = url.path[1:],
            user = url.username,
            password= url.password,
            )

    # fell through connection scheme cases
    raise RuntimeError("{url.scheme} in {db_url} is not supported".format(url=url, db_url=db_url))


class UTA0(Interface0):
    tx_exons_sql = 'select * from tx_exons where ac=? and ref=? order by g_start_i'
    tx_info_sql = 'select * from tx_info where ac=?'
    tx_seq_sql = 'select gene,ac,seq from transcript where ac=?'
    tx_for_gene_sql = 'select * from tx_info where gene=? order by cds_end_i-cds_start_i desc'
    acs_for_protein_md5_sql = 'select ac from protein_hash where md5=?'
        
    def __init__(self):
        self._check_schema_version('0.0')

    def _check_schema_version(self,required_version):
        obs_Mm = self.schema_version().split('.')[:2]
        req_Mm = required_version.split('.')[:2]
        if ( obs_Mm != req_Mm ):
            raise RuntimeError("Version mismatch: Version {req_Mm} required, but {db_path} is version {obs_Mm}".format(
                req_Mm = '.'.join(req_Mm), db_path = self._db_path, obs_Mm = '.'.join(obs_Mm)))
        # else no error

    def data_version(self):
        cur = self._get_cursor()
        cur.execute("select * from meta where key = 'schema_version'")
        return cur.fetchone()['value']

    def schema_version(self):
        return self.data_version()

    def get_tx_exons(self,ac,ref):    
        """
        return transcript info for supplied accession (ac), or None if not found
        
        :param ac: transcript accession with version (e.g., 'NM_000051.3')
        :type ac: str
        :param ref: reference genome ('GRCh37.p10' is the only valid value at this time)
        :type ref: str
        
        # tx_exons = db.get_tx_exons('NM_000051.3','GRCh37.p10')
        # len(tx_exons)
        63
        
        tx_exons have the following attributes::
        
          {'ac': 'NM_000051.3',               # transcript accession
          'ref': 'GRCh37.p10',
          'g_start_i': 108093558,             # genomic start coordinate
          'g_end_i': 108093913,               # genome end coordinate
          'name': '1',
          'ord': 1,
          't_start_i': 0
          't_end_i': 355,
          'g_cigar': '355M',                  # CIGAR string, relative to genome
          'g_seq_a': None,
          't_seq_a': None,
          }
        
        .. note:: chromosome and strand are in the tx_info record
        
        For example:
        
        # tx_exons[0]['ac']
        'NM_000051.3'
        
        """
        cur = self._get_cursor()
        cur.execute(self.tx_exons_sql,[ac,ref])
        return cur.fetchall()

    def get_tx_info(self,ac):
        """return transcript info for supplied accession (ac), or None if not found

        :param ac: transcript accession with version (e.g., 'NM_000051.3')
        :type ac: str
        """
        cur = self._get_cursor()
        cur.execute(self.tx_info_sql,[ac])
        return cur.fetchone()

    def get_tx_seq(self,ac):
        """return transcript sequence for supplied accession (ac), or None if not found

        :param ac: transcript accession with version (e.g., 'NM_000051.3')
        :type ac: str
        """
        cur = self._get_cursor()
        cur.execute(self.tx_seq_sql,[ac])
        try:
            return cur.fetchone()['seq']
        except TypeError:
            return None

    def get_tx_for_gene(self,gene):
        """
        return transcript info records for supplied gene, in order of decreasing length

        :param gene: HGNC gene name
        :type gene: str
        """
        cur = self._get_cursor()
        cur.execute(self.tx_for_gene_sql,[gene])
        return cur.fetchall()

    def get_acs_for_protein_seq(self,seq):
        """
        return a list of protein accessions for a given sequence
        """
        md5 = seq_md5(seq)
        cur = self._get_cursor()
        cur.execute(self.acs_for_protein_md5_sql,[md5])
        return [ r['ac'] for r in cur.fetchall() ]


class UTA0_sqlite(UTA0):
    def __init__(self,db_path):
        def _sqlite3_row_dict_factory(cur, row):
            "convert sqlite row to dict"
            return dict( (d[0],row[i]) for i,d in enumerate(cur.description) )
        if not os.path.exists(db_path):
            raise IOError(db_path + ': Non-existent database file')
        self._con = sqlite3.connect(db_path)
        self._con.row_factory = _sqlite3_row_dict_factory
        logger.info("connected to "+db_path)
        super(UTA0_sqlite,self).__init__()
        
    def _get_cursor(self):
        return self._con.cursor()


class UTA0_postgresql(UTA0):
    def __init__(self,host,port,database,user,password):
        logger.info('connecting to to (host={host}, port={port}, database={database}, user={user}, password={pw})...'.format(
            host=host,port=port,database=database,user=user,pw='' if password is None else '***'))
        self._con = psycopg2.connect(host=host, port=port, database=database, user=user, password=password)
        super(UTA0_postgresql,self).__init__()
        
        self.tx_exons_sql = self.tx_exons_sql.replace('?','%s')
        self.tx_info_sql = self.tx_info_sql.replace('?','%s')
        self.tx_seq_sql = self.tx_seq_sql.replace('?','%s')
        self.tx_for_gene_sql = self.tx_for_gene_sql.replace('?','%s')
        self.acs_for_protein_md5_sql = self.acs_for_protein_md5_sql.replace('?','%s')

        
    def _get_cursor(self):
        cur = self._con.cursor(cursor_factory=psycopg2.extras.DictCursor)
        return cur


if __name__ == '__main__':
    import bdi.sources.uta0
    u0_sl = bdi.sources.uta0.connect('sqlite:///tmp/uta-0.0.5.db')
    u0_pg = bdi.sources.uta0.connect('postgresql://localhost/uta')
    
    cur_sl = u0_sl._get_cursor()
    cur_pg = u0_pg._get_cursor()



## <LICENSE>
## Copyright 2014 BDI Contributors (https://bitbucket.org/invitae/bdi)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>
